
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.fileprocessing;

import static de.unkrig.commons.lang.ExceptionUtil.wrap;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.zip.ZipEntry;

import de.unkrig.commons.file.ExceptionHandler;
import de.unkrig.commons.file.contentsprocessing.ContentsProcessor;
import de.unkrig.commons.file.contentsprocessing.GzipContentsProcessor;
import de.unkrig.commons.file.contentsprocessing.ZipContentsProcessor;
import de.unkrig.commons.file.zipentryprocessing.SelectiveZipEntryProcessor;
import de.unkrig.commons.file.zipentryprocessing.ZipEntryContentsProcessor;
import de.unkrig.commons.file.zipentryprocessing.ZipEntryProcessor;
import de.unkrig.commons.lang.protocol.Predicate;

public
class FileProcessingUtil {

    private FileProcessingUtil() {}

    /**
     * @see #process(List, FileProcessor, ExceptionHandler)
     */
    public static void
    process(List<File> files, FileProcessor fileProcessor) throws IOException {
        process(files, fileProcessor, ExceptionHandler.<IOException>defaultHandler());
    }

    /**
     * Invokes the {@code fileProcessor} for each of the {@code files}.
     *
     * @param exceptionHandler Called if the processing of one of the {@code inputFiles} throws an {@link IOException}
     *                         or a {@link RuntimeException}
     */
    public static void
    process(
        List<File>                    files,
        FileProcessor                 fileProcessor,
        ExceptionHandler<IOException> exceptionHandler
    ) throws IOException {
        for (File file : files) {
            try {
                fileProcessor.process(file);
            } catch (IOException ioe) {
                exceptionHandler.handle(wrap("Processing '" + file + "'", ioe));
            } catch (RuntimeException re) {
                exceptionHandler.handle(wrap("Processing '" + file + "'", re));
            }
        }
    }

    /**
     * @see #newZipFileProcessor(boolean, Predicate, ContentsProcessor, ExceptionHandler)
     */
    public static FileProcessor
    newZipFileProcessor(
        boolean           lookIntoZip,
        Predicate<String> pathPredicate,
        ContentsProcessor contentsProcessor
    ) {
        return newZipFileProcessor(
            lookIntoZip,
            pathPredicate,
            contentsProcessor,
            ExceptionHandler.<IOException>defaultHandler()
        );
    }

    /**
     * Returns a {@link FileProcessor} that calls the {@code contentsProcessor} for every file that matches the
     * {@code pathPredicate}, except if {@code lookIntoZip} is {@code true} and the file is in ZIP or GZIP format. In
     * the latter case, the {@code contentsProcessor} in invoked for each entry in the ZIP file, except for those which
     * are in ZIP format; these are processed recursively.
     * <pre>
     *                 |
     *                 v
     * +-----+  ZipFileProcessor
     * |     |   |entry     |default
     * |     v   v          v
     * | SelectiveZEP     FileCP
     * |       |            |
     * |       v            |
     * |   ZipEntryCP       |
     * |       |            |
     * |       v            v
     * |     GzipCP       GzipCP
     * |       |   contents| |default
     * |       | +---------+ |
     * |       v v           |
     * |      ZipCP          |
     * |  entry| |default    |
     * +-------+ +---------+ |
     *                     | |
     *                     v v
     *                  defaultCP
     * </pre>
     */
    public static FileProcessor
    newZipFileProcessor(
        boolean                       lookIntoZip,
        Predicate<String>             pathPredicate,
        ContentsProcessor             defaultContentsProcessor,
        ExceptionHandler<IOException> exceptionHandler
    ) {
        if (!lookIntoZip) return new FileContentsProcessor(defaultContentsProcessor);

        final ZipEntryProcessor zipEntryProcessor[] = new ZipEntryProcessor[1];

        ContentsProcessor contentsProcessor = new ZipContentsProcessor(
            pathPredicate,            // nestedZipPathPredicate
            new ZipEntryProcessor() { // nestedZipEntryProcessor

                @Override public void
                process(ZipEntry entry, String path, InputStream inputStream) throws IOException {
                    zipEntryProcessor[0].process(entry, path, inputStream);
                }
            },
            defaultContentsProcessor, // defaultContentsProcessor
            exceptionHandler          // exceptionHandler
        );

        FileProcessor fileProcessor = new FileContentsProcessor(new GzipContentsProcessor(contentsProcessor, defaultContentsProcessor));

        contentsProcessor = new GzipContentsProcessor(contentsProcessor, contentsProcessor);

        zipEntryProcessor[0] = new ZipEntryContentsProcessor(contentsProcessor);

        zipEntryProcessor[0] = new SelectiveZipEntryProcessor(pathPredicate, zipEntryProcessor[0], ZipEntryProcessor.NOP);

        fileProcessor = new ZipFileProcessor(
            zipEntryProcessor[0],
            fileProcessor,      // defaultFileProcessor
            exceptionHandler
        );

        return fileProcessor;
    }

    /**
     * @see #newDirectoryTreeProcessor(Predicate, FileProcessor, ExceptionHandler)
     */
    public static FileProcessor
    newDirectoryTreeProcessor(
        Predicate<String> defaultFileProcessor,
        FileProcessor     fileProcessor
    ) {
        return newDirectoryTreeProcessor(
            defaultFileProcessor,
            fileProcessor,
            ExceptionHandler.<IOException>defaultHandler()
        );
    }

    /**
     * Creates a file processor that recurses through the directory tree, applies the {@code directoryPathPredicate},
     * and eventually calls the delegate file processor.
     * <pre>
     *          |
     *          |   +---------+
     *          |   |         |
     *          v   v         |
     *       SelectiveFP      |
     *            |           |
     *            v           |
     *   DirectoryProcessor   |
     *   default|  |member    |
     *          |  +----------+
     *          v
     *    fileProcessor
     * </pre>
     */
    public static FileProcessor
    newDirectoryTreeProcessor(
        Predicate<String>             pathPredicate,
        FileProcessor                 fileProcessor,
        ExceptionHandler<IOException> exceptionHandler
    ) {
        final FileProcessor[] memberProcessor = new FileProcessor[1];

        memberProcessor[0] = new DirectoryProcessor(
            new FileProcessor() {   // memberProcessor
                @Override public void process(File file) throws IOException { memberProcessor[0].process(file); }
            },
            fileProcessor,          // defaultFileprocessor
            exceptionHandler        // exceptionHandler
        );

        memberProcessor[0] = new SelectiveFileProcessor(pathPredicate, memberProcessor[0], FileProcessor.NOP);

        return memberProcessor[0];
    }
}
