
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.fileprocessing;

import static de.unkrig.commons.lang.ExceptionUtil.wrap;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import de.unkrig.commons.file.ExceptionHandler;
import de.unkrig.commons.file.zipentryprocessing.ZipEntryProcessor;

/**
 * @see #process(File)
 */
public
class ZipFileProcessor implements FileProcessor {

    private final ZipEntryProcessor             zipEntryProcessor;
    private final FileProcessor                 defaultFileProcessor;
    private final ExceptionHandler<IOException> exceptionHandler;

    public
    ZipFileProcessor(
        ZipEntryProcessor zipEntryProcessor,
        FileProcessor     defaultFileProcessor
    ) {
        this(
            zipEntryProcessor,
            defaultFileProcessor,
            ExceptionHandler.<IOException>defaultHandler()
        );
    }

    public
    ZipFileProcessor(
        ZipEntryProcessor             zipEntryProcessor,
        FileProcessor                 defaultFileProcessor,
        ExceptionHandler<IOException> exceptionHandler
    ) {
        this.zipEntryProcessor = zipEntryProcessor;
        this.defaultFileProcessor = defaultFileProcessor;
        this.exceptionHandler = exceptionHandler;
    }

    /**
     * If {@code lookIntoZip} is {@code false}, or if the {@code file} is not in ZIP format, or it cannot be opened as
     * a ZIp file, then the {@code file} is passed to the {@code defaultFileProcessor}. Otherwise, if the path of
     * the {@code file}, augmented with '!', does not match the {@code zipFilePredicate}, nothing is done.
     * Otherwise, the {@code file} is opened as a ZIP file, and each ZIP file entry is passed to the {@code
     * zipEntryProcessor}.
     */
    @Override public void
    process(File file) throws FileNotFoundException, IOException {

        // Attempt to process the input file as a ZIP file.

        // Using "new ZipFile(file)" to check whether the file is in ZIP format is terribly expensive. For
        // performance reasons, we check ourselves by looking at the leading magic number.
        boolean isZip;
        {
            InputStream is = new FileInputStream(file);
            try {
                byte[] buf = new byte[4];
                isZip = (
                    is.read(buf) == 4
                    && buf[3] == 0x04 && buf[2] == 0x03 && buf[1] == 0x4B && buf[0] == 0x50
                );
            } catch (IOException ioe) {
                this.exceptionHandler.handle(wrap("Reading the first few bytes from '" + file + "'", ioe));
                return;
            } catch (RuntimeException re) {
                this.exceptionHandler.handle(wrap("Reading the first few bytes from '" + file + "'", re));
                return;
            } finally {
                try { is.close(); } catch (Exception ioe) {}
            }
        }

        if (!isZip) {
            this.defaultFileProcessor.process(file);
            return;
        }

        ZipFile zipFile = new ZipFile(file);
        try {

            String prefix = zipFile.getName() + '!';
            for (Enumeration<? extends ZipEntry> entries = zipFile.entries(); entries.hasMoreElements();) {
                ZipEntry entry = entries.nextElement();

                this.zipEntryProcessor.process(entry, prefix + entry.getName(), zipFile.getInputStream(entry));
            }
        } finally {
            try { zipFile.close(); } catch (Exception e) {}
        }
    }
}
