
/*
 * de.unkrig.loggifier - Injects logging code into JAVA[TM] class files
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.loggifier;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import de.unkrig.commons.file.ExceptionHandler;
import de.unkrig.commons.file.contentstransformation.ContentsTransformer;
import de.unkrig.commons.file.contentstransformation.ContentsTransformerUtil;
import de.unkrig.commons.file.contentstransformation.SelectiveContentsTransformer;
import de.unkrig.commons.file.filetransformation.FileAdderUtil;
import de.unkrig.commons.file.filetransformation.FileTransformationUtil;
import de.unkrig.commons.file.filetransformation.FileTransformer;
import de.unkrig.commons.file.zipentrytransformation.ZipEntriesTransformer;
import de.unkrig.commons.file.zipentrytransformation.ZipEntryTransformer;
import de.unkrig.commons.lang.protocol.Predicate;
import de.unkrig.commons.lang.protocol.PredicateUtil;
import de.unkrig.commons.text.pattern.Glob;
import de.unkrig.loggifier.LoggifyingClassAdapter.ErrorHandler;
import de.unkrig.loggifier.loglevel.LevelCalculator;

/**
 * A {@link FileTransformer} that traverses directory trees, traverses ZIP files, and then loggifies selected
 * class files.
 */
public
class Loggifier implements FileTransformer {

    private final FileTransformer fileTransformer;

    LoggifyingContentsTransformer loggifyingContentsTransformer;

    public Loggifier(
        Predicate<String>  pathSelector,
        LevelCalculator    logLevelCalculator,
        boolean            keepOriginals,
        final ErrorHandler errorHandler
    ) {

        // Set up the loggifying contents transformer.
        this.loggifyingContentsTransformer = new LoggifyingContentsTransformer(logLevelCalculator, errorHandler);

        // Wrap the contents transformer so that only the contents of selected files are loggified.
        ContentsTransformer contentsTransformer = new SelectiveContentsTransformer(
            pathSelector,
            this.loggifyingContentsTransformer,
            ContentsTransformerUtil.copy()
        );

        ExceptionHandler<IOException> ioExceptionHandler = new ExceptionHandler<IOException>() {
            @Override public void handle(IOException ioe)     { errorHandler.handle(ioe); }
            @Override public void handle(RuntimeException re) { errorHandler.handle(re); }
        };

        ZipEntriesTransformer zipEntriesTransformer = new ZipEntriesTransformer() {

            @Override public void
            visitEntry(
                String              prefix,
                ZipEntry            entry,
                InputStream         is,
                ZipEntryTransformer zipEntryTransformer,
                ZipOutputStream     out
            ) throws IOException {
                String entryName = entry.getName();
                zipEntryTransformer.transform(entry, prefix + entryName, is, out);
            }

            @Override public void
            visitEnd(String prefix, ZipOutputStream out) {
            }
        };

        // Wrap the contents transformer in a ZIP file transformer.
        FileTransformer fileTransformer = FileTransformationUtil.newZipFileTransformer(
            true,                                          // lookIntoZip
            PredicateUtil.<String>never(),                 // zipEntryRemoval
            Glob.NONE,                                     // zipEntryRenaming
            zipEntriesTransformer,                         // zipEntriesTransformer
            keepOriginals,                                 // keepOriginals
            contentsTransformer,                           // defaultContentsTransformer
            ioExceptionHandler                             // exceptionHandler
        );

        // Re-wrap the ZIP file transformer in a directory tree transformer.
        fileTransformer = FileTransformationUtil.newDirectoryTreeTransformer(
            PredicateUtil.<String>never(), // fileRemoval
            Glob.NONE,                     // fileRenaming
            FileAdderUtil.nop(),           // fileAdder
            keepOriginals,
            fileTransformer,
            ioExceptionHandler
        );

        this.fileTransformer = fileTransformer;
    }

    @Override public void
    transform(File in, File out) throws IOException {
        this.fileTransformer.transform(in, out);
    }
}
