
/*
 * de.unkrig.loggifier - Injects logging code into JAVA[TM] class files
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.loggifier;

import java.io.IOException;
import java.io.InputStream;

import de.unkrig.commons.io.IoUtil;
import de.unkrig.commons.lang.ExceptionUtil;
import de.unkrig.commons.nullanalysis.NotNullByDefault;
import de.unkrig.commons.nullanalysis.Nullable;
import de.unkrig.loggifier.LoggifyingClassAdapter.ErrorHandler;
import de.unkrig.loggifier.loglevel.LevelCalculator;

/**
 * A class loader that loggifies classes on-the-fly while they are loaded.
 * <p>
 * Usage example:
 * <pre>
 *     ClassLoader loggifyingClassLoader = new LoggifyingClassLoader(
 *         ClassLoader.getSystemClassLoader()         // Delegate is the CLASSPATH loader
 *     );
 *     loggifyingClassLoader.loadClass("pkg.MyClass");
 * </pre>
 */
@NotNullByDefault(false) public final
class LoggifyingClassLoader extends ClassLoader {

    private final ClassLoader                   delegate;
    private final LoggifyingContentsTransformer loggifyingContentsTransformer;

    /**
     * Throws {@link RuntimeException}s with a {@link ClassNotFoundException} cause.
     */
    private static final ErrorHandler CNFE_ERROR_HANDLER = new ErrorHandler() {

        @NotNullByDefault(false) @Override public void
        handle(@Nullable Throwable t) {
            throw new RuntimeException(new ClassNotFoundException().initCause(t));
        }

        @NotNullByDefault(false) @Override public void
        handle(@Nullable String message) {
            throw new RuntimeException(new ClassNotFoundException(message));
        }
    };

    /**
     * Creates a {@link LoggifyingClassLoader} with default log levels that loggifies all classes through the system
     * class loader.
     */
    public
    LoggifyingClassLoader() {
        this(
            ClassLoader.getSystemClassLoader().getParent(),
            ClassLoader.getSystemClassLoader(),
            LoggifyingContentsTransformer.DEFAULT_LOG_LEVEL_CALCULATOR,
            CNFE_ERROR_HANDLER
        );
    }

    /**
     * Creates a {@link LoggifyingClassLoader} with default log levels.
     *
     * @param delegate Used to retrieve the CLASS file resources which are then loggified.
     */
    public
    LoggifyingClassLoader(ClassLoader delegate) {
        this(
            delegate.getParent(),
            delegate,
            LoggifyingContentsTransformer.DEFAULT_LOG_LEVEL_CALCULATOR,
            CNFE_ERROR_HANDLER
        );
    }

    /**
     * Creates a {@link LoggifyingClassLoader} that loads CLASS files through a given {@code delegate} class loader.
     *
     * @param delegate Used to retrieve the CLASS file resources which are then loggified.
     */
    public LoggifyingClassLoader(
        ClassLoader     parent,
        ClassLoader     delegate,
        LevelCalculator logLevelCalculator,
        ErrorHandler    errorHandler
    ) {
        super(parent);
        this.delegate = delegate;
        this.loggifyingContentsTransformer = new LoggifyingContentsTransformer(
            logLevelCalculator,
            errorHandler
        );
    }

    @Override protected Class<?>
    findClass(String className) throws ClassNotFoundException {

        InputStream is = this.delegate.getResourceAsStream(className.replace('.', '/') + ".class");
        if (is == null) throw new ClassNotFoundException(className);

        // Read and transform the class file contents.
        byte[] bytes;
        if (className.startsWith("de.unkrig.loggifier.runtime.")) {
            try {
                bytes = IoUtil.readAll(is);
            } catch (IOException ioe) {
                throw ExceptionUtil.wrap(className, ioe, ClassNotFoundException.class);
            }
        } else {
            try {
                bytes = this.loggifyingContentsTransformer.transform(className, is);
                if (bytes == null) throw new ClassFormatError();
            } catch (MethodCodeTooLargeException mctle) {
                System.err.println("Loggifying '" + className + "': " + mctle);
                try {
                    bytes = IoUtil.readAll(this.delegate.getResourceAsStream(className.replace('.', '/') + ".class"));
                } catch (IOException ioe) {
                    throw ExceptionUtil.wrap(className, ioe, ClassNotFoundException.class);
                }
            } catch (RuntimeException re) {
                Throwable cause = re.getCause();
                if (cause instanceof ClassNotFoundException) throw (ClassNotFoundException) cause;
                throw re;
            } catch (IOException ioe) {
                throw ExceptionUtil.wrap(className, ioe, ClassNotFoundException.class);
            } finally {
                try { is.close(); } catch (Exception e) {}
            }
        }

        // Call "defineClass()".
        return this.defineClass(className, bytes, 0, bytes.length);
    }
}